from .file_storage import FileStorage, ResourceId
from .models.base import Base
from .models.chat import (
    Chat,
    Context,
    ChatMessage,
    AIMessage,
    HumanMessage,
    MiscMessage,
    make_content,
    Content,
    BaseMessage,
)
from .models.chat_item import ChatItem, OuterContextItem, InnerContextItem, ReplicaItem
from .models.enums import MTRSLabelEnum, DiagnosticsXMLTagEnum, MTRSXMLTagEnum, DoctorChoiceXMLTagEnum
from .models.tracks import TrackInfo, DomainInfo
from .models.widget import Widget
from .utils import make_session_id, chunked
from .xml_parser import XMLParser
from .utils_import import load_main_objects

__all__ = [
    "AIMessage",
    "Base",
    "BaseMessage",
    "Chat",
    "ChatItem",
    "ChatMessage",
    "Content",
    "Context",
    "DiagnosticsXMLTagEnum",
    "DoctorChoiceXMLTagEnum",
    "DomainInfo",
    "FileStorage",
    "HumanMessage",
    "InnerContextItem",
    "MTRSLabelEnum",
    "MTRSXMLTagEnum",
    "MiscMessage",
    "OuterContextItem",
    "ReplicaItem",
    "ResourceId",
    "TrackInfo",
    "Widget",
    "XMLParser",
    "chunked",
    "load_main_objects",
    "make_content",
    "make_session_id",
]



from typing import Any, Union, Annotated, get_args, get_origin, Literal

from pydantic import Discriminator, BaseModel, Tag


class _TypeUnionMeta(type):
    def __getitem__(cls, types):
        if not isinstance(types, tuple):
            types = (types,)

        for tp in types:
            if not isinstance(tp, type) or not issubclass(tp, BaseModel):
                raise ValueError(f"Type {tp} must derived from BaseModel")

        # Create tagged union
        tagged_types = []
        type_map = {}
        type_names = set()
        for t in types:
            type_annot = t.__annotations__.get("type")
            if not type_annot or get_origin(type_annot) != Literal:
                raise ValueError(f"Type {t} must have a 'type' Literal[..] field for discrimination")
            type_name = get_args(type_annot)[0]
            type_map[type_name] = t
            tagged_types.append(Annotated[t, Tag(type_name)])
            type_names.add(type_name)

        # Create the union type
        union_type = Union[tuple(tagged_types)]

        # Create discriminator function
        def type_discriminator(v: Any) -> str:
            if isinstance(v, types):
                return v.type
            if isinstance(v, dict):
                tp = v.get("type")
                return tp if tp in type_names else None
            return None

        # Add discriminator
        return Annotated[union_type, Discriminator(type_discriminator)]


class TypeUnion(metaclass=_TypeUnionMeta):
    """
    Wrapper around Union which derive type from field 'type' for effective deserialization via TypeAdapter.

    Usage example:
    ```
    class BaseFruit(BaseModel):
        type: str

    class Apple(BaseFruit):
        type: Literal['apple'] = 'apple'

    class Orange(BaseFruit):
        type: Literal['orange'] = 'orange'

    Fruit = TypeUnion[Apple, Orange]
    ```
    """

    pass



import json
import string
from datetime import datetime
from hashlib import md5
from pathlib import Path
from zipfile import ZipFile, is_zipfile

ResourceId = str
ASCII_DIGITS_SPECIAL = set(string.ascii_lowercase + string.digits + "-")
SUFFIX_DIR = ".dir"
SUFFIX_METADATA = ".metadata"


def _validate_exist(files_dir):
    if not files_dir.exists():
        err = f"Failed to access file-storage directory: {files_dir}"
        raise OSError(err)


def _validate_dtype(dtype: str):
    if all(map(ASCII_DIGITS_SPECIAL.__contains__, dtype)):
        return
    raise ValueError(f"Bad dtype: {dtype}")


def generate_fname(content, dtype):
    fname_hash = md5(content).hexdigest()
    fname = f"{fname_hash}.{dtype}"
    return fname


class FileStorage:
    def __init__(self, files_dir):
        self.files_dir = Path(files_dir)
        self.files_dir.mkdir(exist_ok=True, parents=True)
        _validate_exist(self.files_dir)

    def _generate_fname_path(self, content: bytes, dtype: str):
        fpath = self.files_dir / generate_fname(content, dtype)
        return fpath

    def upload_maybe(self, content: bytes | str | None, fname: str) -> ResourceId | None:
        if not content:
            return None
        resource_id = self.upload(content, fname)
        return resource_id

    def upload(self, content: bytes | str, fname: str) -> ResourceId:
        if isinstance(content, str):
            content = content.encode()

        dtype = fname.rsplit(".", 1)[-1]
        _validate_dtype(dtype)
        fpath = self._generate_fname_path(content, dtype)
        fpath.write_bytes(content)

        fpath_md = fpath.with_suffix(SUFFIX_METADATA)
        update_date = f"{datetime.now():%Y-%m-%d--%H-%M-%S}"
        metadata = {"fname": fname, "update_date": update_date, "size": len(content)}
        fpath_md.write_text(json.dumps(metadata, ensure_ascii=False))

        return str(fpath)

    def get_metadata(self, resource_id: ResourceId) -> dict | None:
        metadata_path = Path(resource_id).with_suffix(SUFFIX_METADATA)
        if not metadata_path.exists():
            return None
        return json.loads(metadata_path.read_text())

    async def upload_async(self, content: bytes | str, fname: str) -> ResourceId:
        return self.upload(content, fname)

    def upload_dir(self, resource_ids: list[ResourceId]) -> ResourceId:
        content = "\n".join(resource_ids)
        res = self.upload(content, "dir")
        return res

    def download(self, resource_id: ResourceId) -> bytes:
        return Path(resource_id).read_bytes()

    async def download_async(self, resource_id: ResourceId) -> bytes:
        return self.download(resource_id)

    def download_text(self, resource_id: ResourceId) -> str:
        return Path(resource_id).read_text(encoding="utf-8")

    def read_dir_or_none(self, resource_id: ResourceId) -> list[ResourceId] | None:
        if not self.is_dir(resource_id):
            return None
        res = self.download_text(resource_id).split("\n")
        return res

    def _get_path(self, resource_id: ResourceId | None) -> Path | None:
        if not resource_id:
            return None
        path = Path(resource_id)
        return path if (path.exists() and path.is_file()) else None

    def is_valid(self, resource_id: ResourceId | None) -> bool:
        path = self._get_path(resource_id)
        return path is not None

    def is_file(self, resource_id: ResourceId | None) -> bool:
        path = self._get_path(resource_id)
        return path and path.suffix != SUFFIX_DIR

    def is_dir(self, resource_id: ResourceId | None) -> bool:
        path = self._get_path(resource_id)
        return path and path.suffix == SUFFIX_DIR

    def get_dtype(self, resource_id: ResourceId | None) -> str | None:
        return resource_id and resource_id.rsplit(".")[-1]

    def unzip_file(self, resource_id: str) -> ResourceId:
        """takes resource_id which refer to zip-archive, unpacks it and returns directory ResourceId with content of zip-archive"""
        path = self._get_path(resource_id)
        if not path:
            raise ValueError(f"Not found path: {resource_id}")
        if not is_zipfile(resource_id):
            raise ValueError(f"Expected zip archive but found: {resource_id}")

        resource_ids = []

        with ZipFile(path, mode="r") as zip_file:
            for file_info in zip_file.filelist:
                file_dtype = file_info.filename.rsplit(".")[-1]
                file_bytes = zip_file.read(file_info)
                rid = self.upload(file_bytes, file_dtype)
                resource_ids.append(rid)

        res = self.upload_dir(resource_ids)
        return res



from pkgutil import iter_modules
from importlib import import_module
from types import ModuleType

from loguru import logger


def convert_snake_to_pascal(name: str) -> str:
    """snake_to_pascal -> SnakeToPascal"""
    return "".join(map(str.capitalize, name.split("_")))


def get_main_object_name(module: type[ModuleType]) -> str:
    module_name = module.__name__.rsplit(".", 1)[-1]
    main_object_name = convert_snake_to_pascal(module_name)
    return main_object_name


def get_main_object(module: type[ModuleType], obj_type):
    """tries to find SomeObject in object src.some_object"""
    main_object_name = get_main_object_name(module)
    try:
        res = getattr(module, main_object_name)
        if not isinstance(res, obj_type) and not issubclass(res, obj_type):
            logger.error(f"Failed to load {module}.{main_object_name}: expected {obj_type} but found {type(res)}")
            return None
        return res
    except AttributeError as ex:
        logger.error(f"Failed to load {module}.{main_object_name}: {ex}")
        return None


def load_modules(package_name: str) -> list[type[ModuleType]]:
    try:
        package = import_module(package_name)
    except ModuleNotFoundError:
        logger.error(f"Not found module: {package_name}")
        return []
    res = [import_module(module_name) for _, module_name, _ in iter_modules(package.__path__, package_name + ".")]
    return res


def load_main_objects(package_name: str, obj_type: type) -> dict[str, object]:
    modules = load_modules(package_name)
    main_objects = [get_main_object(m, obj_type) for m in modules]
    res = {obj.__name__: obj for obj in main_objects}
    return res



from datetime import datetime
from typing import Annotated, Any
from collections.abc import Callable

from mmar_mapi.models.widget import Widget
from pydantic import Field, ConfigDict, BeforeValidator, AfterValidator

from .base import Base


_DT_FORMAT: str = "%Y-%m-%d-%H-%M-%S"
_EXAMPLE_DT_0 = datetime(1970, 1, 1, 0, 0, 0)
_EXAMPLE_DT: str = _EXAMPLE_DT_0.strftime(_DT_FORMAT)


def now_pretty() -> str:
    return datetime.now().strftime(ReplicaItem.DATETIME_FORMAT())


class OuterContextItem(Base):
    # remove annoying warning for protected `model_` namespace
    model_config = ConfigDict(protected_namespaces=())

    sex: bool = Field(False, alias="Sex", description="True = male, False = female", examples=[True])
    age: int = Field(0, alias="Age", examples=[20])
    user_id: str = Field("", alias="UserId", examples=["123456789"])
    parent_session_id: str | None = Field(None, alias="ParentSessionId", examples=["987654320"])
    session_id: str = Field("", alias="SessionId", examples=["987654321"])
    client_id: str = Field("", alias="ClientId", examples=["543216789"])
    track_id: str = Field(default="Consultation", alias="TrackId")
    entrypoint_key: str = Field("", alias="EntrypointKey", examples=["giga"])
    language_code: str = Field("ru", alias="LanguageCode", examples=["ru"])

    def create_id(self, short: bool = False) -> str:
        uid, sid, cid = self.user_id, self.session_id, self.client_id
        if short:
            return f"{uid}_{sid}_{cid}"
        return f"user_{uid}_session_{sid}_client_{cid}"

    def to_dict(self) -> dict[str, Any]:
        return self.model_dump(by_alias=True)


LABELS = {
    0: "OK",
    1: "NON_MED",
    2: "CHILD",
    3: "ABSURD",
    4: "GREETING",
    5: "RECEIPT",
}


def fix_deprecated_moderation(moderation):
    if isinstance(moderation, int):
        return LABELS.get(moderation, "OK")
    elif isinstance(moderation, str):
        return moderation
    else:
        raise ValueError(f"Unsupported moderation: {moderation} :: {type(moderation)}")


def nullify_empty(text: str) -> str | None:
    return text or None


class ReplicaItem(Base):
    body: str = Field("", alias="Body", examples=["Привет"])
    resource_id: Annotated[str | None, AfterValidator(nullify_empty)] = Field(
        None, alias="ResourceId", examples=["<link-id>"]
    )
    resource_name: Annotated[str | None, AfterValidator(nullify_empty)] = Field(
        None, alias="ResourceName", examples=["filename"]
    )
    widget: Widget | None = Field(None, alias="Widget", examples=[None])
    command: dict | None = Field(None, alias="Command", examples=[None])
    role: bool = Field(False, alias="Role", description="True = ai, False = client", examples=[False])
    date_time: str = Field(
        default_factory=now_pretty, alias="DateTime", examples=[_EXAMPLE_DT], description=f"Format: {_DT_FORMAT}"
    )
    state: str = Field("", alias="State", description="chat manager fsm state", examples=["COLLECTION"])
    action: str = Field("", alias="Action", description="chat manager fsm action", examples=["DIAGNOSIS"])
    # todo fix: support loading from `moderation: int`
    moderation: Annotated[str, BeforeValidator(str)] = Field(
        "OK", alias="Moderation", description="moderation outcome", examples=["OK"]
    )
    extra: dict | None = Field(None, alias="Extra", examples=[None])

    def to_dict(self) -> dict[str, Any]:
        return self.model_dump(by_alias=True)

    @staticmethod
    def DATETIME_FORMAT() -> str:
        return _DT_FORMAT

    def with_now_datetime(self):
        return self.model_copy(update=dict(date_time=now_pretty()))

    @property
    def is_ai(self):
        return self.role

    @property
    def is_human(self):
        return not self.role

    def modify_text(self, callback: Callable[[str], str]) -> "ReplicaItem":
        body_upd = callback(self.body)
        return self.model_copy(update=dict(body=body_upd))


class InnerContextItem(Base):
    replicas: list[ReplicaItem] = Field(alias="Replicas")
    attrs: dict[str, str | int] | None = Field(default={}, alias="Attrs")

    def to_dict(self) -> dict[str, list]:
        return self.model_dump(by_alias=True)


class ChatItem(Base):
    outer_context: OuterContextItem = Field(alias="OuterContext")
    inner_context: InnerContextItem = Field(alias="InnerContext")

    def create_id(self, short: bool = False) -> str:
        return self.outer_context.create_id(short)

    def to_dict(self) -> dict[str, Any]:
        return self.model_dump(by_alias=True)

    def add_replica(self, replica: ReplicaItem):
        self.inner_context.replicas.append(replica)

    def add_replicas(self, replicas: list[ReplicaItem]):
        for replica in replicas:
            self.inner_context.replicas.append(replica)

    def replace_replicas(self, replicas: list[ReplicaItem]):
        return self.model_copy(update=dict(inner_context=InnerContextItem(replicas=replicas)))

    def get_last_state(self, default: str = "empty") -> str:
        replicas = self.inner_context.replicas
        for ii in range(len(replicas) - 1, -1, -1):
            replica = replicas[ii]
            if replica.role:
                return replica.state
        return default

    def zip_history(self, field: str) -> list[Any]:
        return [replica.to_dict().get(field, None) for replica in self.inner_context.replicas]

    @classmethod
    def parse(cls, chat_obj: str | dict) -> "ChatItem":
        return _parse_chat_item(chat_obj)


def _parse_chat_item(chat_obj: str | dict) -> ChatItem:
    if isinstance(chat_obj, dict):
        return ChatItem.model_validate(chat_obj)

    return ChatItem.model_validate_json(chat_obj)
